==> pmml-evaluator-1.4.0-sources.jar: org/jpmml/evaluator/ArrayUtil.java <==
/*
 * Copyright (c) 2012 University of Tartu
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.jpmml.evaluator;

import java.util.ArrayList;
import java.util.List;

import com.google.common.base.Function;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Interner;
import com.google.common.collect.Interners;
import com.google.common.collect.Lists;
import org.dmg.pmml.Array;

public class ArrayUtil {

	private ArrayUtil(){
	}

	static
	public int getSize(Array array){
		Integer n = array.getN();
		if(n != null){
			return n;
		}

		List<?> content = getContent(array);

		return content.size();
	}

	static
	public List<?> getContent(Array array){
		return CacheUtil.getValue(array, ArrayUtil.contentCache);
	}

	static
	public List<? extends Number> asNumberList(Array array){
		List<?> content = getContent(array);

		Array.Type type = array.getType();
		if(type == null){
			throw new MissingAttributeException(array, PMMLAttributes.ARRAY_TYPE);
		}

		switch(type){
			case INT:
			case REAL:
				return (List)content;
			case STRING:
				throw new InvalidElementException(array);
			default:
				throw new UnsupportedAttributeException(array, type);
		}
	}

	static
	public List<String> tokenize(Array array){
		List<String> result;

		Array.Type type = array.getType();
		if(type == null){
			throw new MissingAttributeException(array, PMMLAttributes.ARRAY_TYPE);
		}

		String value = array.getValue();

		switch(type){
			case INT:
			case REAL:
				result = tokenize(value, false);
				break;
			case STRING:
				result = tokenize(value, true);
				break;
			default:
				throw new UnsupportedAttributeException(array, type);
		}

		Integer n = array.getN();
		if(n != null && n != result.size()){
			throw new InvalidElementException(array);
		}

		return result;
	}

	static
	public List<?> parse(Array array){
		List<String> tokens = tokenize(array);

		Array.Type type = array.getType();
		if(type == null){
			throw new MissingAttributeException(array, PMMLAttributes.ARRAY_TYPE);
		}

		switch(type){
			case INT:
				return Lists.transform(tokens, INT_PARSER);
			case REAL:
				return Lists.transform(tokens, REAL_PARSER);
			case STRING:
				return Lists.transform(tokens, STRING_PARSER);
			default:
				throw new UnsupportedAttributeException(array, type);
		}
	}

	static
	private List<String> tokenize(String string, boolean enableQuotes){
		List<String> result = new ArrayList<>();

		StringBuilder sb = new StringBuilder();

		boolean quoted = false;

		tokens:
		for(int i = 0; i < string.length(); i++){
			char c = string.charAt(i);

			if(quoted){

				if(c == '\\' && i < (string.length() - 1)){
					c = string.charAt(i + 1);

					if(c == '\"'){
						sb.append('\"');

						i++;
					} else

					{
						sb.append('\\');
					}

					continue tokens;
				} // End if

				sb.append(c);

				if(c == '\"'){
					result.add(createToken(sb, enableQuotes));

					quoted = false;
				}
			} else

			{
				if(c == '\"' && enableQuotes){

					if(sb.length() > 0){
						result.add(createToken(sb, enableQuotes));
					}

					sb.append('\"');

					quoted = true;
				} else

				if(Character.isWhitespace(c)){

					if(sb.length() > 0){
						result.add(createToken(sb, enableQuotes));
					}
				} else

				{
					sb.append(c);
				}
			}
		}

		if(sb.length() > 0){
			result.add(createToken(sb, enableQuotes));
		}

		return result;
	}

	static
	private String createToken(StringBuilder sb, boolean enableQuotes){
		String result;

		if(sb.length() > 1 && (sb.charAt(0) == '\"' && sb.charAt(sb.length() - 1) == '\"') && enableQuotes){
			result = sb.substring(1, sb.length() - 1);
		} else

		{
			result = sb.substring(0, sb.length());
		}

		sb.setLength(0);

		return result;
	}

	private static final Interner<Integer> integerInterner = Interners.newWeakInterner();
	private static final Interner<Double> doubleInterner = Interners.newWeakInterner();
	private static final Interner<String> stringInterner = Interners.newWeakInterner();

	private static final Function<String, Integer> INT_PARSER = new Function<String, Integer>(){

		@Override
		public Integer apply(String string){
			return ArrayUtil.integerInterner.intern(Integer.parseInt(string));
		}
	};

	private static final Function<String, Double> REAL_PARSER = new Function<String, Double>(){

		@Override
		public Double apply(String string){
			return ArrayUtil.doubleInterner.intern(Double.parseDouble(string));
		}
	};

	private static final Function<String, String> STRING_PARSER = new Function<String, String>(){

		@Override
		public String apply(String string){
			return ArrayUtil.stringInterner.intern(string);
		}
	};

	private static final LoadingCache<Array, List<?>> contentCache = CacheUtil.buildLoadingCache(new CacheLoader<Array, List<?>>(){

		@Override
		public List<?> load(Array array){
			return ImmutableList.copyOf(parse(array));
		}
	});
}

==> pmml-evaluator-1.4.1-sources.jar: org/jpmml/evaluator/ArrayUtil.java <==
/*
 * Copyright (c) 2018 Villu Ruusmann
 *
 * This file is part of JPMML-Evaluator
 *
 * JPMML-Evaluator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JPMML-Evaluator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with JPMML-Evaluator.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.jpmml.evaluator;

import java.util.List;

import com.google.common.base.Function;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import org.dmg.pmml.Array;

public class ArrayUtil {

	private ArrayUtil(){
	}

	static
	public int getSize(Array array){
		Integer n = array.getN();
		if(n != null){
			return n;
		}

		List<?> content = getContent(array);

		return content.size();
	}

	static
	public List<?> getContent(Array array){
		return CacheUtil.getValue(array, ArrayUtil.contentCache);
	}

	static
	public List<? extends Number> asNumberList(Array array){
		List<?> content = getContent(array);

		Array.Type type = array.getType();
		if(type == null){
			throw new MissingAttributeException(array, PMMLAttributes.ARRAY_TYPE);
		}

		switch(type){
			case INT:
			case REAL:
				return (List)content;
			case STRING:
				throw new InvalidElementException(array);
			default:
				throw new UnsupportedAttributeException(array, type);
		}
	}

	static
	public List<?> parse(Array array){
		Array.Type type = array.getType();
		if(type == null){
			throw new MissingAttributeException(array, PMMLAttributes.ARRAY_TYPE);
		}

		String value = array.getValue();

		List<String> tokens;

		switch(type){
			case INT:
			case REAL:
				tokens = org.jpmml.model.ArrayUtil.parse(value, false);
				break;
			case STRING:
				tokens = org.jpmml.model.ArrayUtil.parse(value, true);
				break;
			default:
				throw new UnsupportedAttributeException(array, type);
		}

		Integer n = array.getN();
		if(n != null && n != tokens.size()){
			throw new InvalidElementException(array);
		}

		switch(type){
			case INT:
				return Lists.transform(tokens, INT_PARSER);
			case REAL:
				return Lists.transform(tokens, REAL_PARSER);
			case STRING:
				return Lists.transform(tokens, STRING_PARSER);
			default:
				throw new UnsupportedAttributeException(array, type);
		}
	}

	private static final Function<String, Integer> INT_PARSER = new Function<String, Integer>(){

		@Override
		public Integer apply(String string){
			return Numbers.INTEGER_INTERNER.intern(Integer.parseInt(string));
		}
	};

	private static final Function<String, Double> REAL_PARSER = new Function<String, Double>(){

		@Override
		public Double apply(String string){
			return Numbers.DOUBLE_INTERNER.intern(Double.parseDouble(string));
		}
	};

	private static final Function<String, String> STRING_PARSER = new Function<String, String>(){

		@Override
		public String apply(String string){
			return Strings.INTERNER.intern(string);
		}
	};

	private static final LoadingCache<Array, List<?>> contentCache = CacheUtil.buildLoadingCache(new CacheLoader<Array, List<?>>(){

		@Override
		public List<?> load(Array array){
			return ImmutableList.copyOf(parse(array));
		}
	});
}
